@@ -3,8 +3,10 @@ import logging
 import os
 import platform
 import subprocess
+from datetime import date
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from threading import Lock
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, Literal
 
 import flet
 from flet import (
@@ -17,22 +19,29 @@ from flet import (
     FilePicker,
     FilePickerResultEvent,
     Image,
+    MainAxisAlignment,
     Markdown,
     Radio,
     RadioGroup,
+    Row,
     Text,
     TextField,
+    alignment,
     colors,
     icons,
 )
 from flet.control_event import ControlEvent
 from flet.dropdown import Option
 from robot.api.deco import keyword, library
+from robot.errors import RobotError
 from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
-from RPA.Assistant.flet_client import FletClient
 
-from .dialog_types import Icon, Options, Result, Size
-from .utils import optional_str, to_options
+from RPA.Assistant.flet_client import FletClient
+from RPA.Assistant.types import Icon, Options, Result, Size, Location
+from RPA.Assistant.utils import (
+    optional_str,
+    to_options,
+)
 
 
 @library(scope="GLOBAL", doc_format="REST", auto_keywords=False)
@@ -139,11 +148,14 @@ class Assistant:
         self.logger = logging.getLogger(__name__)
         self._client = FletClient()
         self._validations: Dict[str, Callable] = {}
+        # Disable fletd debug logging
+        os.environ["GIN_MODE"] = "release"
 
         try:
             # Prevent logging from keywords that return results
             keywords = [
                 "Run dialog",
+                "Ask user",
             ]
             BuiltIn().import_library(
                 "RPA.core.logger.RobotLogListener", "WITH NAME", "RPA.RobotLogListener"
@@ -154,31 +166,33 @@ class Assistant:
         except RobotNotRunningError:
             pass
 
-    def _add_closing_button(self, label="Submit") -> None:
+    def _create_closing_button(self, label="Submit") -> Control:
         def validate_and_close(*_):
             should_close = True
             if not self._validations:
                 self._client.page.window_destroy()
             for field_name, validation in self._validations.items():
-                if not validation:
+                if not validation or field_name not in self._client.results:
                     continue
                 error_message = validation(self._client.results[field_name])
                 if error_message:
                     should_close = False
-                    self._client.page.add(Text(
-                        f"Error on field named '{field_name}: {error_message}",
-                        color=flet.colors.RED))
+                    self._client.page.add(
+                        Text(
+                            f"Error on field named '{field_name}: {error_message}",
+                            color=flet.colors.RED,
+                        )
+                    )
                     self._client.page.update()
             if should_close:
                 self._client.results["submit"] = label
                 self._client.page.window_destroy()
 
-        self._client.add_element(ElevatedButton(label, on_click=validate_and_close))
+        return ElevatedButton(label, on_click=validate_and_close)
 
     @keyword("Clear elements")
-    def clear_elements(self) -> None:
-        """Remove all previously defined elements and start from a clean state
-
+    def clear_dialog(self) -> None:
+        """Remove all previously defined elements and start from a clean state.
         By default this is done automatically when a dialog is created.
 
         Example:
@@ -192,9 +206,10 @@ class Assistant:
             END
             Clear elements
         """
+        self._client.results = {}
         self._client.clear_elements()
 
-    @keyword("Add heading")
+    @keyword("Add Heading")
     def add_heading(
         self,
         heading: str,
@@ -287,7 +302,12 @@ class Assistant:
         """
         if not label:
             label = url
-        self._client.add_element(Markdown(f"[{label}]({url})"))
+        self._client.add_element(
+            Markdown(
+                f"[{label}]({url})",
+                on_tap_link=lambda e: self._client.page.launch_url(e.data),
+            )
+        )
 
     @keyword("Add image")
     def add_image(
@@ -467,13 +487,14 @@ class Assistant:
         name: str,
         label: Optional[str] = None,
         placeholder: Optional[str] = None,
-        validation: Union[Callable, str] = None
+        validation: Union[Callable, None] = None,
     ) -> None:
         """Add a text input element
 
         :param name:        Name of result field
         :param label:       Label for field
         :param placeholder: Placeholder text in input field
+        :param validation   Validation function for the input field
 
         Adds a text field that can be filled by the user. The entered
         content will be available in the ``name`` field of the result.
@@ -494,14 +515,13 @@ class Assistant:
             ${result}=    Run dialog
             Send feedback message    ${result.email}  ${result.message}
         """
-        # TODO: Do this in a cleaner way. Workaround because we use on_change
-        # handlers to record values, so default value otherwise will be missed
         # TODO: Implement the rows support
-        self._client.results[name] = placeholder
-        self._validations[name] = validation
+        # TODO: add robot keyword support e.g. if is keyword name use BuiltIn().run_keyword()
+        if validation:
+            self._validations[name] = validation
 
         self._client.add_element(
-            name=name, element=TextField(label=label, value=placeholder)
+            name=name, element=TextField(label=label, hint_text=placeholder)
         )
 
     @keyword("Add password input", tags=["input"])
@@ -600,8 +620,8 @@ class Assistant:
 
         The argument ``file_type`` restricts the possible file extensions
         that the user can select. The format of the argument is as follows:
-        ``Description text (*.ext1;*.ext2;...)``. For instance, an argument
-        to limit options to Excel files could be: ``Excel files (*.xls;*.xlsx)``.
+        ``pdf,png,svg``. For instance, an argument
+        to limit options to Excel files could be: ``xls,xlsx``.
 
         To allow selecting more than one file, the ``multiple`` argument
         can be enabled.
@@ -623,7 +643,7 @@ class Assistant:
             Add file input    name=dest      destination=%{ROBOT_ROOT}
 
             # This restricts files to certain types
-            Add file input    name=types     file_type=PDF files (*.pdf)
+            Add file input    name=types     file_type=pdf
 
             # Every file input result is a list of paths
             ${result}=    Run dialog
@@ -648,6 +668,9 @@ class Assistant:
         if not options["source"]:
             options["source"] = os.path.expanduser("~")
 
+        if options["file_type"]:
+            options["file_type"] = options["file_type"].split(",")
+
         self._client.add_element(
             ElevatedButton(
                 label or "Choose files...",
@@ -705,36 +728,57 @@ class Assistant:
         self._client.add_element(Text(value=label))
         self._client.add_element(dropdown, name=str(name))
 
-    # FIXME: Add keyword back when fixed
-    # @keyword("Add Date Input", tags=["input"])
-    # def add_date_input(
-    #     self,
-    #     name: str,
-    #     default: Optional[Union[date, str]] = None,
-    #     label: Optional[str] = None,
-    # ) -> None:
-    #     """Add a date input element
+    @keyword("Add Date Input", tags=["input"])
+    def add_date_input(
+        self,
+        name: str,
+        default: Optional[Union[date, str]] = None,
+        label: Optional[str] = None,
+    ) -> None:
+        """Add a date input element.
 
-    #     :param name:    Name of the result field
-    #     :param default: The default set date
-    #     :param label:   Label for the date input field
+        :param name:    Name of the result field
+        :param default: The default set date
+        :param label:   Label for the date input field
 
-    #     Displays a date input widget. The selection the user makes will be available
-    #     as a ``date`` object in the ``name`` field of the result.
-    #     The ``default`` argument can be a pre-set date as object or string in
-    #     "YYYY-MM-DD" format, otherwise the current date is used.
+        Displays a date input widget. The selection the user makes will be available
+        as a ``date`` object in the ``name`` field of the result.
+        The ``default`` argument can be a pre-set date as object or string in
+        "YYYY-MM-DD" format, otherwise the current date is used.
 
-    #     Example:
+        Example:
 
-    #     .. code-block:: robotframework
+        .. code-block:: robotframework
 
-    #         Add heading       Enter your birthdate
-    #         Add Date Input    birthdate    default=1993-04-26
-    #         ${result} =       Run dialog
-    #         Log To Console    User birthdate year should be: ${result.birthdate.year}
-    #     """
+            Add heading       Enter your birthdate
+            Add Date Input    birthdate    default=1993-04-26
+            ${result} =       Run dialog
+            Log To Console    User birthdate year should be: ${result.birthdate.year}
+        """
 
-    #     self._client.add_element(name=str(name), element=DatePicker())
+        def validate(date_text):
+            try:
+                date.fromisoformat(date_text)
+                return None
+            except ValueError:
+                return "Date should be in format YYYY-MM-DD"
+
+        if default:
+            if isinstance(default, str):
+                try:
+                    default = date.fromisoformat(default)
+                except ValueError as e:
+                    self.logger.error(
+                        f"Default date value {default} is not in a valid ISO format."
+                    )
+                    raise e
+            self._client.results[name] = default
+        # TODO: add robot keyword support e.g. if is keyword name use BuiltIn().run_keyword()
+        self._validations[name] = validate
+        self._client.add_element(
+            name=name,
+            element=TextField(label=label, hint_text="YYYY-MM-DD", value=default),
+        )
 
     @keyword("Add radio buttons", tags=["input"])
     def add_radio_buttons(
@@ -773,7 +817,9 @@ class Assistant:
             Log    User type should be: ${result.user_type}
         """
         options, default = to_options(options, default)
-        radios: List[Control] = [Radio(value=option, label=option) for option in options]
+        radios: List[Control] = [
+            Radio(value=option, label=option) for option in options
+        ]
         radio_group = RadioGroup(content=Column(radios), value=default)
 
         self._client.add_element(Text(value=label))
@@ -818,9 +864,7 @@ class Assistant:
 
     @keyword("Add submit buttons", tags=["input"])
     def add_submit_buttons(
-        self,
-        buttons: Options,
-        default: Optional[str] = None
+        self, buttons: Options, default: Optional[str] = None
     ) -> None:
         """Add custom submit buttons
 
@@ -854,18 +898,27 @@ class Assistant:
                 Delete user    ${username}
             END
         """
-        buttons, default = to_options(buttons, default)
-        for button in buttons:
-            self._add_closing_button(button)
+        button_labels, default = to_options(buttons, default)
+
+        button_elements = [
+            self._create_closing_button(button) for button in button_labels
+        ]
+        for button in button_elements:
+            self._client.add_to_disablelist(button)
+
+        button_row = Row(button_elements, alignment=MainAxisAlignment.END)
+        container = Container(button_row, alignment=alignment.bottom_right)
+        self._client.add_element(container)
 
     @keyword("Run dialog", tags=["dialog"])
     def run_dialog(
         self,
         timeout: int = 180,
         title: str = "Dialog",
-        height: Union[int, str] = "AUTO",
+        height: Union[int, Literal["AUTO"]] = "AUTO",
         width: int = 480,
         on_top: bool = False,
+        location: Union[Location, Tuple[int, int], None] = None,
     ) -> Result:
         """Create a dialog from all the defined elements and block
         until the user has handled it.
@@ -875,6 +928,8 @@ class Assistant:
         :param height: Height of dialog (in pixels or 'AUTO')
         :param width:  Width of dialog (in pixels)
         :param on_top: Show dialog always on top of other windows
+        :param location: Where to place the dialog (options are Center, TopLeft, or a tuple of ints)
+        None will let the operating system place the window.
 
 
         Returns a result object with all input values.
@@ -889,9 +944,18 @@ class Assistant:
             Log    The username is: ${result.username}
         """
 
-        # FIXME: support timeout
+        # height has to be either AUTO or an int value
+        if not isinstance(height, int) and height != "AUTO":
+            height = int(height)
 
-        self._client.display_flet_window(title, height, width, on_top)
+        # if location is given as a string (Robot autoconversion doesn't work) parse it
+        # to enum manually
+        if isinstance(location, str):
+            location = Location[location]
+
+        self._client.display_flet_window(
+            title, height, width, on_top, location, timeout
+        )
         return self._client.results
 
     @keyword("Ask User", tags=["dialog"])
@@ -918,11 +982,8 @@ class Assistant:
             Log    The username is: ${result.username}
         """
 
-        # FIXME: support timeout
-
         self.add_submit_buttons(["Submit", "Close"], "Submit")
-        self._client.display_flet_window(**options)
-        return self._client.results
+        return self.run_dialog(**options, timeout=timeout)
 
     @keyword("Refresh Dialog", tags=["dialog"])
     def refresh_dialog(self):
@@ -932,42 +993,87 @@ class Assistant:
         else:
             raise RuntimeError("No dialog open")
 
+    def _queue_function_or_robot_keyword(
+        self, function: Union[Callable, str], *args, **kwargs
+    ):
+        """Check if function is a Python function or a Robot Keyword, and call it
+        or run it with Robot's run_keyword.
+        """
+        if self._client.pending_operation:
+            self.logger.error(f"Can't have more than one pending operation.")
+            return
+        self._client.lock_elements()
+        self._client.flet_update()
+
+        if isinstance(function, Callable):
+
+            def func_wrapper():
+                try:
+                    function(*args, **kwargs)
+
+                # This can be anything since it comes from the user function, we don't
+                # want to let the user function crash the UI
+                # pylint: disable=W0703
+                except Exception as e:
+                    self.logger.error(f"Error calling Python function {function}")
+                    self.logger.error(e)
+                finally:
+                    self._client.unlock_elements()
+                    self._client.flet_update()
+
+            self._client.pending_operation = func_wrapper
+        else:
+
+            def func_wrapper():
+                try:
+                    BuiltIn().run_keyword(function, *args, **kwargs)
+                except RobotNotRunningError:
+                    self.logger.error(
+                        f"Robot Framework not running so cannot call keyword {function}"
+                    )
+                except RobotError as e:
+                    self.logger.error(f"Error calling robot keyword {function}")
+                    self.logger.error(e)
+                # This can be anything since it comes from the user function, we don't
+                # want to let the user function crash the UI
+                # pylint: disable=W0703
+                except Exception as e:
+                    self.logger.error(
+                        f"Unexpected error running robot keyword {function}"
+                    )
+                    self.logger.error(e)
+                finally:
+                    self._client.unlock_elements()
+                    self._client.flet_update()
+
+            self._client.pending_operation = func_wrapper
+
     @keyword("Add Button", tags=["dialog"])
     def add_button(
         self, label: str, function: Union[Callable, str], *args, **kwargs
     ) -> None:
         """
-        ``function`` should be a python function or a Robot keyword name,
-        args and kwargs should be valid arguments for it.
+        :param label Text for the button
+        :param function Python function or Robot Keyword name, that will get *args and
+        **kwargs passed into it
         """
 
-        # FIXME: add some progress bar
-        # TODO: either optional or mandatory feature to block other button function
-        # calls while one is being processed
-        # (perhaps by making the button disabled during execution?)
+        def on_click(_: ControlEvent):
+            self._queue_function_or_robot_keyword(function, *args, **kwargs)
 
-        # TODO: use logger.err and logger.debug
-        def on_click(_):
-            if isinstance(function, Callable):
-                try:
-                    function(*args, **kwargs)
-                except Exception as err:
-                    print(f"on_click error with button labeled {label}")
-                    print(err)
-            else:
-                try:
-                    BuiltIn().run_keyword(function, *args, **kwargs)
-                except Exception as err:
-                    print(f"on_click error with button labeled {label}")
-                    print(err)
-
-        self._client.add_element(ElevatedButton(label, on_click=on_click))
+        button = ElevatedButton(label, on_click=on_click)
+        self._client.add_element(button)
+        self._client.add_to_disablelist(button)
 
     @keyword("Add Next Ui Button", tags=["dialog"])
     def add_next_ui_button(self, label: str, function: Union[Callable, str]):
         """Create a button that leads to the next UI page, calling the passed
         keyword or function, and passing current form results as first positional argument to it.
 
+        :param label Text for the button
+        :param function Python function or Robot Keyword name, that will take form
+        results as it's first argument
+
         Example:
             *** Keywords ***
             Retrieve User Data
@@ -988,18 +1094,9 @@ class Assistant:
 
         """
 
-        def on_click(_):
-            if isinstance(function, Callable):
-                try:
-                    function(self._client.results)
-                except Exception as err:
-                    print(f"on_click error with button labeled {label}")
-                    print(err)
-            else:
-                try:
-                    BuiltIn().run_keyword(function, self._client.results)
-                except Exception as err:
-                    print(f"on_click error with button labeled {label}")
-                    print(err)
+        def on_click(_: ControlEvent):
+            self._queue_function_or_robot_keyword(function, self._client.results)
 
-        self._client.add_element(ElevatedButton(label, on_click=on_click))
+        button = ElevatedButton(label, on_click=on_click)
+        self._client.add_element(button)
+        self._client.add_to_disablelist(button)
